from dune.perftool.generation import (backend,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      temporary_variable,
                                      )

from dune.perftool.sumfact.amatrix import (quadrature_points_per_direction,
                                           name_oned_quadrature_points,
                                           name_oned_quadrature_weights,
                                           )
from dune.perftool.pdelab.argument import name_accumulation_variable
from dune.perftool.pdelab.geometry import (dimension_iname,
                                           local_dimension,
                                           )

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )

import numpy


def nest_quadrature_loops(kernel, inames):
    insns = []
    for insn in kernel.instructions:
        if "quad" in insn.tags:
            insns.append(insn.copy(within_inames=insn.within_inames.union(frozenset(inames)),
                                   tags=insn.tags - frozenset({"quad"})
                                   )
                         )
        else:
            insns.append(insn)

    return kernel.copy(instructions=insns)


class BaseWeight(FunctionIdentifier):
    def __init__(self, accumobj):
        self.accumobj = accumobj

    def __getinitargs__(self):
        return (self.accumobj,)

    @property
    def name(self):
        return '{}.weight'.format(self.accumobj)


@function_mangler
def base_weight_function_mangler(target, func, dtypes):
    if isinstance(func, BaseWeight):
        return CallMangleInfo(func.name, (NumpyType(numpy.float64),), ())


def pymbolic_base_weight():
    """ This is the base weight that should be multiplied to the quadrature
    weight. With the fast DG assembler this will handle the weighting of the
    time discretization scheme.
    TODO: Introduce backend switch that uses above BaseWeight function
    """
    return 1.0


@iname
def sumfact_quad_iname(d, context):
    name = "quad_{}_{}".format(context, d)
    domain(name, quadrature_points_per_direction())
    return name


@backend(interface="quad_inames", name="sumfact")
def quadrature_inames(context=''):
    return tuple(sumfact_quad_iname(d, context) for d in range(local_dimension()))


def define_recursive_quadrature_weight(name, dir):
    iname = quadrature_inames()[dir]
    temporary_variable(name, shape=(), shape_impl=())
    instruction(expression=Product((recursive_quadrature_weight(dir + 1),
                                    Subscript(Variable(name_oned_quadrature_weights()),
                                              (Variable(iname),)
                                              ))
                                   ),
                assignee=Variable(name),
                forced_iname_deps=frozenset(quadrature_inames()[dir:]),
                forced_iname_deps_is_final=True,
                tags=frozenset({"quad"}),
                )


def recursive_quadrature_weight(dir=0):
    if dir == local_dimension():
        return pymbolic_base_weight()
    else:
        name = 'weight_{}'.format(dir)
        define_recursive_quadrature_weight(name, dir)
        return Variable(name)


def quadrature_weight():
    return recursive_quadrature_weight()


def define_quadrature_position(name):
    for i in range(local_dimension()):
        instruction(expression=Subscript(Variable(name_oned_quadrature_points()), (Variable(quadrature_inames()[i]),)),
                    assignee=Subscript(Variable(name), (i,)),
                    forced_iname_deps=frozenset(quadrature_inames()),
                    forced_iname_deps_is_final=True,
                    tags=frozenset({"quad"}),
                    )


@backend(interface="quad_pos", name="sumfact")
def pymbolic_quadrature_position():
    name = 'pos'
    temporary_variable(name, shape=(local_dimension(),), shape_impl=("fv",))
    define_quadrature_position(name)
    return Variable(name)


@backend(interface="qp_in_cell", name="sumfact")
def pymbolic_quadrature_position_in_cell(restriction):
    from dune.perftool.pdelab.geometry import to_cell_coordinates
    return to_cell_coordinates(pymbolic_quadrature_position(), restriction)
